# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import getpass
import json
import os
from time import time

from iconcommons import IconConfig
from iconcommons.logger.logger import Logger
from iconsdk.builder.call_builder import CallBuilder
from iconsdk.builder.transaction_builder import TransactionBuilder, CallTransactionBuilder, DeployTransactionBuilder, \
    MessageTransactionBuilder, DepositTransactionBuilder
from iconsdk.exception import KeyStoreException, DataTypeException
from iconsdk.icon_service import IconService
from iconsdk.providers.http_provider import HTTPProvider
from iconsdk.signed_transaction import SignedTransaction
from iconsdk.utils import add_0x_prefix
from iconsdk.utils.convert_type import convert_hex_str_to_int, convert_bytes_to_hex_str
from iconsdk.utils.templates import BLOCK_0_3_VERSION
from iconsdk.wallet.wallet import KeyWallet
from iconservice.base.address import is_icon_address_valid

from tbears.config.tbears_config import FN_CLI_CONF, tbears_cli_config, keystore_test1, TBEARS_CLI_TAG
from tbears.tbears_exception import TBearsCommandException, JsonContentsException
from tbears.util.arg_parser import uri_parser
from tbears.util.argparse_type import IconAddress, IconPath, hash_type, non_negative_num_type, loop
from tbears.util.keystore_manager import validate_password
from tbears.util.transaction_logger import call_with_logger, send_transaction_with_logger


class CommandWallet:
    def __init__(self, subparsers):
        self._add_txresult_parser(subparsers)
        self._add_transfer_parser(subparsers)
        self._add_keystore_parser(subparsers)
        self._add_keyinfo_parser(subparsers)
        self._add_balance_parser(subparsers)
        self._add_totalsupply_parser(subparsers)
        self._add_scoreapi_parser(subparsers)
        self._add_txbyhash_parser(subparsers)
        self._add_lastblock_parser(subparsers)
        self._add_blockbyhash_parser(subparsers)
        self._add_blockbyheight_parser(subparsers)
        self._add_sendtx_parser(subparsers)
        self._add_call_parser(subparsers)
        self._add_block_parser(subparsers)

    @staticmethod
    def _add_block_parser(subparsers):
        parser = subparsers.add_parser('block',
                                       help="Get block info using given argument."
                                            "Only one of `number` and `hash` arguments can be passed.")
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-i', '--id', dest='hash', type=hash_type, help='Hash of the block to be queried.')
        parser.add_argument('-n', '--number', dest='number', type=non_negative_num_type,
                            help='Height of the block to be queried.')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF}). This command returns the block as it was stored.')

    @staticmethod
    def _add_lastblock_parser(subparsers):
        parser = subparsers.add_parser('lastblock', help='Get last block\'s info', description='Get last block\'s info')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF}). This command returns the block in v0.1a format')

    @staticmethod
    def _add_blockbyhash_parser(subparsers):
        parser = subparsers.add_parser('blockbyhash', help='Get block info using given block hash',
                                       description='Get block info using given block hash')
        parser.add_argument('hash', type=hash_type, help='Hash of the block to be queried.')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF}). This command returns the block in v0.1a format')

    @staticmethod
    def _add_blockbyheight_parser(subparsers):
        parser = subparsers.add_parser('blockbyheight', help='Get block info using given block height',
                                       description='Get block info using given block height')
        parser.add_argument('height', type=non_negative_num_type, help='height of the block to be queried.')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF}). This command returns the block in v0.1a format')

    @staticmethod
    def _add_txresult_parser(subparsers):
        parser = subparsers.add_parser('txresult', help='Get transaction result by transaction hash',
                                       description='Get transaction result by transaction hash')
        parser.add_argument('hash', type=hash_type, help='Hash of the transaction to be queried.')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_transfer_parser(subparsers):
        parser = subparsers.add_parser('transfer', help='Transfer ICX coin.', description='Transfer ICX coin.')
        parser.add_argument('-f', '--from', type=IconAddress(), help='From address.')
        parser.add_argument('to', type=IconAddress(), help='Recipient')
        parser.add_argument("value", type=loop, help='Amount of ICX coin in loop to transfer (1 icx = 1e18 loop)')
        parser.add_argument('-k', '--key-store', type=IconPath(), dest='keyStore',
                            help='Keystore file path. Used to generate "from" address and transaction signature')
        parser.add_argument('-n', '--nid', help='Network ID (default: 0x3)')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-p', '--password', help='Keystore file\'s password', dest='password')
        parser.add_argument('-s', '--step-limit', dest='stepLimit', type=non_negative_num_type, help='Step limit')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default values for '
                                 f'"keyStore", "uri", "from" and "stepLimit". (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_keystore_parser(subparsers):
        parser = subparsers.add_parser('keystore',
                                       help='Create a keystore file in the specified path',
                                       description='Create keystore file in the specified path. Generate privatekey, '
                                                   'publickey pair using secp256k1 library.')
        parser.add_argument('path', type=IconPath('w'), help='Path of keystore file.')
        parser.add_argument('-p', '--password', help='Keystore file\'s password', dest='password')

    @staticmethod
    def _add_keyinfo_parser(subparsers):
        parser = subparsers.add_parser('keyinfo',
                                       help='Show a keystore information in the specified path',
                                       description="Show a keystore information(address, privateKey, publicKey) "                                                
                                                   "in the specified path. If you want to get privateKey, "
                                                   "input --private-key option"
                                       )
        parser.add_argument('path', type=IconPath(), help='Path of keystore file.')
        parser.add_argument('-p', '--password', help='Keystore file\'s password', dest='password')
        parser.add_argument('--private-key', help='option that whether show privateKey', action='store_true',
                            dest='privateKey')

    @staticmethod
    def _add_balance_parser(subparsers):
        parser = subparsers.add_parser('balance',
                                       help='Get balance of given address in loop unit',
                                       description='Get balance of given address in loop unit')
        parser.add_argument('address', type=IconAddress(), help='Address to query the ICX balance')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_totalsupply_parser(subparsers):
        parser = subparsers.add_parser('totalsupply', help='Query total supply of ICX in loop unit',
                                       description='Query total supply of ICX in loop unit')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_scoreapi_parser(subparsers):
        parser = subparsers.add_parser('scoreapi', help='Get SCORE\'s API using given SCORE address',
                                       description='Get SCORE\'s API using given SCORE address')
        parser.add_argument('address', type=IconAddress('cx'), help='SCORE address to query the API')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath,
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_txbyhash_parser(subparsers):
        parser = subparsers.add_parser('txbyhash', help='Get transaction by transaction hash',
                                       description='Get transaction by transaction hash')
        parser.add_argument('hash', type=hash_type, help='Hash of the transaction to be queried.')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_sendtx_parser(subparsers):
        parser = subparsers.add_parser('sendtx', help='Request icx_sendTransaction with the specified json file and '
                                                      'keystore file. If keystore file is not given, tbears sends a '
                                                      'request as it is in the json file.',
                                       description='Request icx_sendTransaction with the specified json file and '
                                                   'keystore file. If keystore file is not given, tbears sends a '
                                                   'request as it is in the json file.')
        parser.add_argument('json_file', type=IconPath(), help='File path containing icx_sendTransaction content')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-n', '--nid', help='Network ID (default: 0x3)')
        parser.add_argument('-k', '--key-store', dest='keyStore', type=IconPath(),
                            help='Keystore file path. Used to generate "from" address and transaction signature')
        parser.add_argument('-p', '--password', help='Keystore file\'s password', dest='password')
        parser.add_argument('-s', '--step-limit', dest='stepLimit', type=non_negative_num_type, help='Step limit')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default values for '
                            f'"keyStore", "uri", "from" and "stepLimit". (default: {FN_CLI_CONF})')

    @staticmethod
    def _add_call_parser(subparsers):
        parser = subparsers.add_parser('call', help='Request icx_call with the specified json file.',
                                       description='Request icx_call with the specified json file.')
        parser.add_argument('json_file', type=IconPath(), help='File path containing icx_call content')
        parser.add_argument('-u', '--node-uri', dest='uri', help='URI of node (default: http://127.0.0.1:9000/api/v3)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'Configuration file path. This file defines the default value for '
                                 f'"uri" (default: {FN_CLI_CONF})')

    @staticmethod
    def _check_transfer(conf: dict, password: str = None):
        if not is_icon_address_valid(conf['to']):
            raise TBearsCommandException(f'You entered invalid address')

        if conf['to'] == keystore_test1['address']:
            uri: str = conf['uri']
            index = uri.find('127.0.0.1')
            if index == -1 or uri[index + len('127.0.0.1')] != ':':
                raise TBearsCommandException(f'Do not transfer to "test1" account')

        if conf.get('keyStore', None):
            if not password:
                password = getpass.getpass("Input your keystore password: ")

        return password

    @staticmethod
    def _check_keystore(password: str):
        if not password:
            password = getpass.getpass("Input your keystore password: ")
            password_retype = getpass.getpass("Retype your keystore password: ")

            if password != password_retype:
                raise TBearsCommandException("Sorry, passwords do not match. Failed to make keystore file")

        if not validate_password(password):
            raise TBearsCommandException("Password must be at least 8 characters long including alphabet, number, "
                                         "and special character.")
        return password

    @staticmethod
    def _check_keyinfo(password: str):
        if not password:
            password = getpass.getpass("Input your keystore password: ")

        return password

    @staticmethod
    def _check_sendtx(conf: dict, password: str = None):
        if conf.get('keyStore', None):
            if not os.path.exists(conf['keyStore']):
                raise TBearsCommandException(f'There is no keystore file {conf["keyStore"]}')
            if not password:
                password = getpass.getpass("Input your keystore password: ")
        else:
            if not is_icon_address_valid(conf['from']):
                raise TBearsCommandException(f'invalid address: {conf["from"]}')

        return password

    def block(self, conf):
        """Query block with given parameter(height or hash)

        :param conf: block command configuration
        :return: result of query
        """
        uri, version = uri_parser(conf['uri'])
        icon_service, response = IconService(HTTPProvider(uri, version)), None
        hash, number = conf.get('hash'), conf.get('number')
        if hash is not None and number is not None:
            raise TBearsCommandException("Only one of id and number can be passed.")
        if hash is not None:
            response = icon_service.get_block(hash, True, BLOCK_0_3_VERSION)
        if number is not None:
            response = icon_service.get_block(convert_hex_str_to_int(number), True, BLOCK_0_3_VERSION)
        if response is None:
            raise TBearsCommandException("You have to specify block height or block hash")

        if "error" in response:
            print(json.dumps(response, indent=4))
        else:
            print(f"block info : {json.dumps(response, indent=4)}")

        return response

    def lastblock(self, conf):
        """Query last block

        :param conf: lastblock command configuration
        :return: result of query
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_block("latest", True)

        if "error" in response:
            print(json.dumps(response, indent=4))
        else:
            print(f"block info : {json.dumps(response, indent=4)}")

        return response

    def blockbyheight(self, conf):
        """Query block with given height

        :param conf: blockbyheight command configuration
        :return: result of query
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_block(convert_hex_str_to_int(conf['height']), True)

        if "error" in response:
            print('Got an error response')
            print(json.dumps(response, indent=4))
        else:
            print(f"block info : {json.dumps(response, indent=4)}")

        return response

    def blockbyhash(self, conf):
        """Query block with given hash

        :param conf: blockbyhash command configuration
        :return: result of query
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_block(conf['hash'], True)

        if "error" in response:
            print('Got an error response')
            print(json.dumps(response, indent=4))
        else:
            print(f"block info : {json.dumps(response, indent=4)}")

        return response

    def txbyhash(self, conf):
        """Query transaction using given transaction hash.

        :param conf: txbyhash command configuration.
        :return: result of query.
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_transaction(conf['hash'], True)

        if "error" in response:
            print('Got an error response')
            print(f"Can not get transaction \n{json.dumps(response, indent=4)}")
        else:
            print(f"Transaction : {json.dumps(response, indent=4)}")

        return response

    def txresult(self, conf):
        """Query transaction result using given transaction hash.

        :param conf: txresult command configuration.
        :return: result of query.
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_transaction_result(conf['hash'], True)

        if "error" in response:
            print('Got an error response')
            print(f"Can not get transaction \n{json.dumps(response, indent=4)}")
        else:
            print(f"Transaction : {json.dumps(response, indent=4)}")

        return response

    def transfer(self, conf: dict) -> dict:
        """Transfer ICX Coin.

        :param conf: transfer command configuration.
        :return: response of transfer.
        """
        # check value type (must be int), address and keystore file
        # if valid, return user input password
        password = conf.get('password', None)
        password = self._check_transfer(conf, password)

        if password:
            try:
                wallet = KeyWallet.load(conf['keyStore'], password)
                from_ = wallet.get_address()

            except KeyStoreException as e:
                print(e.args[0])
                return None
        else:
            # make dummy wallet
            wallet = KeyWallet.create()
            from_ = conf['from']

        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        transaction = TransactionBuilder() \
            .from_(from_) \
            .to(conf['to']) \
            .value(int(conf['value'])) \
            .nid(convert_hex_str_to_int(conf['nid'])) \
            .timestamp(int(time() * 10 ** 6)) \
            .build()

        if 'stepLimit' not in conf:
            step_limit = icon_service.estimate_step(transaction)
        else:
            step_limit = convert_hex_str_to_int(conf['stepLimit'])

        transaction.step_limit = step_limit

        # Returns the signed transaction object having a signature
        signed_transaction = SignedTransaction(transaction, wallet)

        if not password:
            signed_transaction.signed_transaction_dict['signature'] = 'sig'

        # Sends transaction and return response
        response = send_transaction_with_logger(icon_service, signed_transaction, uri)

        if 'result' in response:
            print('Send transfer request successfully.')
            tx_hash = response['result']
            print(f"transaction hash: {tx_hash}")
        else:
            print('Got an error response')
            print(json.dumps(response, indent=4))

        return response

    def keystore(self, conf: dict):
        """Create a keystore file with the specified path and password.

        :param conf: keystore command configuration
        """
        # check if the given keystore file already exists, and if user input is a valid password
        password = conf.get('password', None)
        password = self._check_keystore(password)

        key_store_content = KeyWallet.create()
        key_store_content.store(conf['path'], password)

        print(f"Made keystore file successfully")

    def keyinfo(self, conf: dict):
        """Show a keystore information with the the specified path and password.

        :param conf: keyinfo command configuration
        """
        password = conf.get('password', None)
        password = self._check_keyinfo(password)

        try:
            wallet = KeyWallet.load(conf['path'], password)

            key_info = {
                "address": wallet.get_address(),
                "publicKey": convert_bytes_to_hex_str(wallet.public_key)
            }

            if conf['privateKey']:
                key_info['privateKey'] = add_0x_prefix(wallet.get_private_key())

            print(json.dumps(key_info, indent=4))

            return key_info

        except KeyStoreException as e:
            print(e.args[0])

    def balance(self, conf: dict):
        """Query icx balance of given address

        :param conf: balance command configuration.
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_balance(conf['address'], True)

        if "error" in response:
            print('Got an error response')
            print(json.dumps(response, indent=4))
        else:
            print(f"balance in hex: {response['result']}")
            print(f"balance in decimal: {int(response['result'], 16)}")

        return response

    @staticmethod
    def totalsupply(conf: dict):
        """Query total supply of ICX

        :param conf: totalsupply command configuration
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_total_supply(True)

        if "error" in response:
            print('Got an error response')
            print(json.dumps(response, indent=4))
        else:
            print(f'Total supply of ICX in hex: {response["result"]}')
            print(f'Total supply of ICX in decimal: {int(response["result"], 16)}')

        return response

    def scoreapi(self, conf):
        """Query score API of given score address.

        :param conf: scoreapi command configuration.
        :return: result of query.
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        response = icon_service.get_score_api(conf['address'], True)

        if "error" in response:
            print('Got an error response')
            print(f"Can not get {conf['address']}'s API\n{json.dumps(response, indent=4)}")
        else:
            print(f"SCORE API: {json.dumps(response['result'], indent=4)}")

        return response

    def sendtx(self, conf: dict):
        """Send transaction.

        :param conf: sendtx command configuration.
        :return: response of transfer.
        """
        with open(conf['json_file'], 'r') as jf:
            payload = json.load(jf)

        password = conf.get('password', None)
        password = self._check_sendtx(conf, password)
        params = payload['params']

        if password and conf.get('keyStore'):
            try:
                wallet = KeyWallet.load(conf['keyStore'], password)
                params['from'] = wallet.get_address()

            except KeyStoreException as e:
                print(e.args[0])
                return None
        else:
            # make dummy wallet
            wallet = KeyWallet.create()

        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        transaction = self.get_transaction(conf, params)
        if 'stepLimit' not in conf:
            step_limit = icon_service.estimate_step(transaction)
        else:
            step_limit = convert_hex_str_to_int(conf['stepLimit'])

        transaction.step_limit = step_limit

        signed_transaction = SignedTransaction(transaction, wallet)

        if not password:
            signed_transaction.signed_transaction_dict['signature'] = 'sig'

        # Sends transaction
        response = send_transaction_with_logger(icon_service, signed_transaction, uri)

        if 'result' in response:
            print('Send transaction request successfully.')
            tx_hash = response['result']
            print(f"transaction hash: {tx_hash}")
        else:
            print('Got an error response')
            print(json.dumps(response, indent=4))

        return response

    @staticmethod
    def call(conf):
        """Request icx_call

        :param conf: call command configuration.
        :return: response of icx_call
        """
        uri, version = uri_parser(conf['uri'])
        icon_service = IconService(HTTPProvider(uri, version))

        with open(conf['json_file'], 'r') as jf:
            payload = json.load(jf)

        call = CallBuilder()\
            .from_(conf['from'])\
            .to(payload['params']['to'])\
            .method(payload['params']['data']['method'])\
            .params(payload['params']['data'].get('params', None))\
            .build()

        response = call_with_logger(icon_service, call, uri)

        if 'error' in response:
            print(json.dumps(response, indent=4))
        else:
            print(f'response : {json.dumps(response, indent=4)}')

        return response

    def check_command(self, command):
        return hasattr(self, command)

    def run(self, args):
        if not hasattr(self, args.command):
            raise TBearsCommandException(f"Invalid command {args.command}")

        user_input = vars(args)
        conf = self.get_icon_conf(args.command, args=user_input)

        Logger.info(f"Run '{args.command}' command with config: {conf}", TBEARS_CLI_TAG)

        # run command
        return getattr(self, args.command)(conf)

    @staticmethod
    def get_icon_conf(command: str, args: dict = None) -> dict:
        """Load config file using IconConfig instance
        config file is loaded as below priority
        system config -> default config -> user config -> user input config(higher priority)

        :param command: command name (e.g. balance)
        :param args: user input command (converted to dictionary type)
        :return: command configuration
        """
        # load configurations
        conf = IconConfig(FN_CLI_CONF, copy.deepcopy(tbears_cli_config))
        # load config file
        conf.load(config_path=args.get('config', None) if args else None)

        # move command config
        if command in conf:
            conf.update_conf(conf[command])
            del conf[command]

        # load user argument
        if args:
            conf.update_conf(args)

        return conf

    @staticmethod
    def get_transaction(conf: dict, params: dict):
        data_type = params.get('dataType')
        params_data = params.get('data', {})

        try:
            transaction_params = {
                "from_": params['from'],
                "to": params['to'],
                "nid": convert_hex_str_to_int(conf['nid']),
                "value": convert_hex_str_to_int(params.get('value', "0x0"))
            }

            if data_type is None:
                transaction_builder = TransactionBuilder(**transaction_params)
            elif data_type == "call":
                transaction_params['method'] = params_data.get('method')
                transaction_params['params'] = params_data.get('params')
                transaction_builder = CallTransactionBuilder(**transaction_params)
            elif data_type == "deploy":
                transaction_params['content'] = params_data.get('content')
                transaction_params['content_type'] = params_data.get('contentType')
                transaction_params['params'] = params_data.get('params')
                transaction_builder = DeployTransactionBuilder(**params)
            elif data_type == "message":
                transaction_params['data'] = params_data
                transaction_builder = MessageTransactionBuilder(**transaction_params)
            elif data_type == "deposit":
                transaction_params['action'] = params_data.get('action')
                transaction_params['id'] = params_data.get('id')
                transaction_builder = DepositTransactionBuilder(**transaction_params)
            else:
                raise JsonContentsException("Invalid dataType")
            transaction = transaction_builder.build()
        except KeyError:
            raise JsonContentsException("Invalid json content. check json contents")
        except TypeError:
            raise JsonContentsException("Invalid json content. check keys")
        except DataTypeException:
            raise JsonContentsException("Invalid json content. check values")
        else:
            return transaction



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
import asyncio
from contextlib import suppress


class Periodic:
    """
    Class for periodic work in asyncio
    """
    def __init__(self, func: callable, interval: int):
        self.func = func
        self.interval = interval
        self.is_started = False
        self._task = None

    async def start(self):
        """
        Start the periodic work
        :return:
        """
        if not self.is_started:
            self.is_started = True
            # Start task to call func periodically:
            self._task = asyncio.ensure_future(self._run())

    async def stop(self):
        """
        Stop the periodic work
        :return:
        """
        if self.is_started:
            self.is_started = False
            # Stop task and await it stopped:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run(self):
        """
        Do the work
        :return:
        """
        next_time = time.time() + self.interval
        while True:
            # get time to sleep
            remain_time = next_time - time.time()
            if remain_time > 0:
                await asyncio.sleep(remain_time)

            # set next working time
            next_time = time.time() + self.interval

            # do work
            await self.func()


class Immediate:
    """
    Class for immediate work in asyncio
    """
    def __init__(self):
        self.funcs: list = []
        self.is_started = False
        self._task = None

    async def start(self):
        """
        Start the immediate work
        :return:
        """
        if not self.is_started:
            self.is_started = True
            # Start task to call func periodically:
            self._task = asyncio.ensure_future(self._run())

    async def stop(self):
        """
        Stop the immediate work
        :return:
        """
        if self.is_started:
            self.is_started = False
            # Stop task and await it stopped:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    def add_func(self, func: callable):
        self.funcs.append(func)

    async def _run(self):
        """
        Do the work
        :return:
        """
        while True:
            if self.funcs:
                func: callable = self.funcs.pop()
                await func()
            else:
                await asyncio.sleep(0)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import json
import os
import warnings
from functools import wraps
from inspect import getmembers, isfunction
from typing import Optional
from unittest.mock import Mock, patch

from iconservice import InterfaceScore
from iconservice.base.address import Address, AddressPrefix
from iconservice.base.exception import InvalidPayableException, InvalidInterfaceException, InvalidRequestException
from iconservice.database.db import IconScoreDatabase
from iconservice.icon_constant import IconScoreFuncType, IconScoreContextType
from iconservice.iconscore.icon_score_base2 import _create_address_with_key, _recover_key
from iconservice.iconscore.icon_score_constant import CONST_BIT_FLAG, ConstBitFlag, FORMAT_IS_NOT_DERIVED_OF_OBJECT
from iconservice.iconscore.icon_score_context import IconScoreContext, ContextGetter
from iconservice.iconscore.icon_score_context_util import IconScoreContextUtil
from iconservice.iconscore.icx import Icx

from .context import Context, score_mapper, interface_score_mapper
from ..mock.db import MockKeyValueDatabase
from ..mock.icx_engine import IcxEngine

context_db = None


def hooking_get_balance(address: 'Address'):
    """Hooking method for icx.get_balance"""
    return IcxEngine.get_balance(None, address)


def hooking_transfer(_to: 'Address', amount: int):
    """Hooking method for icx.transfer"""
    ctx = ContextGetter._context
    sender = ctx.msg.sender
    IcxEngine.transfer(ContextGetter._context, sender, _to, amount)


def hooking_sha_256(data):
    return hashlib.sha256(data).digest()


def hooking_sha3_256(data):
    return hashlib.sha3_256(data).digest(0)


def start_SCORE_APIs_patch(score_module_path: str):
    patch(f"{score_module_path}.json_dumps", side_effect=json.dumps).start()
    patch(f"{score_module_path}.json_loads", side_effect=json.loads).start()
    patch(f"{score_module_path}.sha_256", side_effect=hooking_sha_256).start()
    patch(f"{score_module_path}.sha3_256", side_effect=hooking_sha3_256).start()
    patch(f"{score_module_path}.create_address_with_key", side_effect=_create_address_with_key).start()
    patch(f"{score_module_path}.recover_key", side_effect=_recover_key).start()


def create_address(prefix: AddressPrefix = AddressPrefix.EOA) -> 'Address':
    return Address.from_bytes(prefix.to_bytes(1, 'big') + os.urandom(20))


def deprecated_method(param):
    warnings.warn("forbidden method", DeprecationWarning, stacklevel=2)


def patch_score_method(method):
    """Patch SCORE method.
    Refer to the decorator and patch the method to have the appropriate context

    :param method: method to patch
    :return: patched method
    """

    @wraps(method)
    def patched(*args, **kwargs):
        context: 'Mock' = ContextGetter._context
        method_flag = getattr(method, CONST_BIT_FLAG, 0)
        score_class, method_name = method.__qualname__.split('.')
        context.current_address = method.__self__.address

        if method_name == 'fallback':
            if not (method_flag & ConstBitFlag.Payable) and context.msg.value > 0:
                raise InvalidPayableException(f"This method is not payable")

        if method_flag & ConstBitFlag.ReadOnly == ConstBitFlag.ReadOnly:
            Context._set_query_context(context)
        else:
            Context._set_invoke_context(context)
        if method_flag & ConstBitFlag.Payable:
            IcxEngine.transfer(context, context.msg.sender, context.current_address, context.msg.value)

        context.readonly = context.type == IconScoreContextType.QUERY or context.func_type == IconScoreFuncType.READONLY

        result = method(*args, **kwargs)
        return result

    return patched


def get_interface_score(score_address):
    try:
        interface_score = interface_score_mapper[score_address]
    except KeyError:
        raise InvalidInterfaceException(FORMAT_IS_NOT_DERIVED_OF_OBJECT.format(InterfaceScore.__name__))
    else:
        return interface_score


def new_create_interface_score(score_address, interface_score):
    """Hooking method for SCORE.create_interface_score

    :param score_address: address of internal call SCORE
    :param interface_score:
    :return: mock instance
    """
    return get_interface_score(score_address)


class ScorePatcher:

    @staticmethod
    def get_score_db(score_address: Optional['Address'] = None):
        """Get db of SCORE that having score_address.
        create cx prefixed address and set it as SCORE's address if score_address is None

        :param score_address: address of score.
        :return: db SCORE use
        """
        global context_db
        if not score_address:
            score_address = create_address(AddressPrefix.CONTRACT)
        score_db = IconScoreDatabase(score_address, context_db)
        return score_db

    @staticmethod
    def initialize_score(score_class, score_db, owner: 'Address'):
        """Get an instance of the SCORE class passed as an score_class arguments

        :param score_class: SCORE class to instantiate
        :param score_db: database the SCORE use
        :param owner: owner of SCORE
        :return: Instantiated SCORE
        """
        context = Context.get_context()
        ScorePatcher._set_mock_context(context)
        Context.set_msg(owner, 0)
        score = score_class(score_db)
        ScorePatcher.patch_score_methods(score)
        ScorePatcher.patch_score_event_logs(score)
        ScorePatcher.patch_interface_scores(score)
        ScorePatcher.patch_deprecated_methods(score)
        score_mapper[score.address] = score
        return score

    @staticmethod
    def _set_mock_context(context: 'IconScoreContext'):
        mock_context = Mock(spec=IconScoreContext)
        mock_context.configure_mock(msg=context.msg)
        mock_context.configure_mock(tx=context.tx)
        mock_context.configure_mock(block=context.block)
        mock_context.configure_mock(step_counter=None)
        mock_context.configure_mock(type=context.type)
        mock_context.configure_mock(func_type=context.func_type)
        mock_context.configure_mock(current_address=context.current_address)
        mock_context.configure_mock(block_batch=context.block_batch)
        mock_context.configure_mock(tx_batch=context.tx_batch)
        mock_context.configure_mock(event_logs=context.event_logs)
        mock_context.configure_mock(event_log_stack=context.event_log_stack)
        mock_context.configure_mock(traces=context.traces)
        mock_context.configure_mock(icon_score_mapper=context.icon_score_mapper)
        ContextGetter._context = mock_context

    @staticmethod
    def patch_score_methods(score):
        """Patch all SCORE method by calling patch_score_method function

        :param score: SCORE to be patched
        """
        custom_methods = ScorePatcher.get_custom_methods(score.__class__)
        for custom_method in custom_methods:
            name = custom_method.__qualname__.split('.')[1]
            method = getattr(score, name)
            setattr(score, name, patch_score_method(method))

    @staticmethod
    def get_custom_methods(score):
        """Get user defined methods inside SCORE

        :param score: SCORE to get the custom method
        :return: user defined methods inside SCORE
        """
        custom_methods = ScorePatcher._get_custom_methods(score)
        methods = set()
        for method in custom_methods:
            if getattr(method, CONST_BIT_FLAG, 0) in (0, 1, 2, 3, 4) or method.__name__ in ('on_install', 'on_update'):
                methods.add(method)

        methods.add(getattr(score, 'fallback'))
        return methods

    @staticmethod
    def _get_custom_methods(score):
        custom_methods = [method for key, method in getmembers(score, predicate=isfunction)
                          if method.__qualname__.split('.')[0] != 'IconScoreBase']
        return custom_methods

    @staticmethod
    def register_interface_score(internal_score_address: Address):
        """Register interface SCORE. This method must be called before testing internal call(Calling other SCORE method)

        :param internal_score_address: address of internal call SCORE
        """
        if not internal_score_address.is_contract:
            raise InvalidRequestException(f"{internal_score_address} is not SCORE")
        interface_score_mapper[internal_score_address] = Mock()

    @staticmethod
    def patch_interface_scores(score):
        """Patch internal call SCORE with mock instance"""
        setattr(score, 'create_interface_score', new_create_interface_score)

    @staticmethod
    def patch_internal_method(internal_score_address, method, new_method):
        """Patch internal method with given 'new_method'

        :param internal_score_address: address of the SCORE having method to be called
        :param method: method to be patched
        :param new_method: method to patch
        """
        interface_score = get_interface_score(internal_score_address)
        setattr(interface_score, method, Mock(side_effect=new_method))

    @staticmethod
    def patch_score_event_logs(score):
        """Patch all event_logs inside SCORE

        :param score: SCORE to be patched
        """
        custom_methods = ScorePatcher._get_custom_methods(score.__class__)
        for method in custom_methods:
            if getattr(method, CONST_BIT_FLAG, 0) == ConstBitFlag.EventLog:
                setattr(score, method.__name__, Mock())

    @staticmethod
    def patch_deprecated_methods(score):
        setattr(score, "get_owner", Mock(side_effect=deprecated_method))
        setattr(score, "is_score_active", Mock(side_effect=deprecated_method))
        setattr(score, "get_score_address_by_tx_hash", Mock(side_effect=deprecated_method))
        setattr(score, "get_tx_hashes_by_score_address", Mock(side_effect=deprecated_method))
        setattr(score, "deploy", Mock(side_effect=deprecated_method))

    @staticmethod
    def start_patches():
        """Patch SCORE to use dictionary DB instance of LEVEL DB"""

        global context_db
        context_db = MockKeyValueDatabase.create_db()
        IcxEngine.db = context_db
        patch('iconservice.iconscore.icon_score_context.ContextGetter._context').start()
        patch.object(Icx, "get_balance", side_effect=hooking_get_balance).start()
        patch.object(Icx, "transfer", side_effect=hooking_transfer).start()
        patch.object(IconScoreContextUtil, "get_owner",
                     side_effect=lambda context, score_address: context.msg.sender).start()

    @staticmethod
    def stop_patches():
        """Stop patching and clear db"""
        patch.stopall()
        IcxEngine.db.close()
        score_mapper.clear()
        interface_score_mapper.clear()



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
import socket
import subprocess
import sys
import time
from ipaddress import ip_address
from typing import Optional

from iconcommons.icon_config import IconConfig
from iconcommons.logger import Logger

from tbears.block_manager.block_manager import TBEARS_BLOCK_MANAGER
from tbears.config.tbears_config import FN_SERVER_CONF, tbears_server_config, ConfigKey, TBEARS_CLI_TAG
from tbears.tbears_exception import TBearsCommandException, TBearsWriteFileException
from tbears.util import write_file
from tbears.util.argparse_type import port_type, IconPath

BLOCKMANAGER_MODULE_NAME = 'tbears.block_manager'
TBEARS_CLI_ENV = '/tmp/.tbears.env'


class CommandServer(object):
    def __init__(self, subparsers):
        self._add_start_parser(subparsers)
        self._add_stop_parser(subparsers)

    @staticmethod
    def _add_start_parser(subparsers) -> None:
        parser = subparsers.add_parser('start', help='Start tbears service',
                                       description='Start tbears service')
        parser.add_argument('-a', '--address', type=ip_address, help='Address to host on (default: 127.0.0.1)', dest='hostAddress')
        parser.add_argument('-p', '--port', type=port_type, help='Port to listen on (default: 9000)')
        parser.add_argument('-c', '--config', type=IconPath(),
                            help=f'tbears configuration file path (default: {FN_SERVER_CONF})')

    @staticmethod
    def _add_stop_parser(subparsers) -> None:
        subparsers.add_parser('stop', help='Stop tbears service',
                              description='Stop all running SCOREs and tbears service')

    def run(self, args):
        if not hasattr(self, args.command):
            raise TBearsCommandException(f"Invalid command {args.command}")

        # load configurations
        user_input = vars(args)
        conf = self.get_icon_conf(args.command, args=user_input)

        Logger.info(f"Run '{args.command}' command with config: {conf}", TBEARS_CLI_TAG)

        # run command
        return getattr(self, args.command)(conf)

    @staticmethod
    def get_icon_conf(command: str, args: dict = None) -> dict:
        """Load config file using IconConfig instance
        config file is loaded as below priority
        system config -> default config -> user config -> user input config (higher priority)

        :param command: command name (e.g. start)
        :param args: user input command (converted to dictionary type)
        :return: command configuration
        """
        # load configurations
        conf = IconConfig(FN_SERVER_CONF, copy.deepcopy(tbears_server_config))
        # load config file
        conf.load(config_path=args.get('config', None) if args else None)

        # move command config
        if command in conf:
            conf.update_conf(conf[command])
            del conf[command]

        # load user argument
        if args:
            conf.update_conf(args)

        # add default configuration file
        if args.get('config', None) is None:
            if os.path.exists(FN_SERVER_CONF):
                conf['config'] = FN_SERVER_CONF

        return conf

    def start(self, conf: dict):
        """ Start tbears service
        Start iconservice, tbears_block_manager, iconrpcserver

        :param conf: start command configuration
        """
        if self.is_service_running():
            raise TBearsCommandException(f"tbears service was started already")

        if self.is_port_available(conf) is False:
            raise TBearsCommandException(f"port {conf['port']} already in use. use other port.")

        # write temporary configuration file
        temp_conf = './temp_conf.json'
        with open(temp_conf, mode='w') as file:
            file.write(json.dumps(conf))

        # run iconservice
        self._start_iconservice(conf, temp_conf)

        # start tbears_block_manager
        self._start_blockmanager(conf)

        # start iconrpcserver
        self._start_iconrpcserver(conf, temp_conf)
        time.sleep(3)

        # remove temporary configuration file
        os.remove(temp_conf)

        # write server configuration
        self.write_server_conf(conf)

        print(f'Started tbears service successfully')

    def stop(self, _conf: dict):
        """ Stop tbears service
        Start iconservice, tbears_block_manager, iconrpcserver

        :param _conf: stop command configuration
        """
        if not self.is_service_running():
            print(f'tbears service is not running')
            return

        # stop iconrpcserver
        subprocess.run(['iconrpcserver', 'stop'], stdout=subprocess.DEVNULL)

        # stop tbears_block_manager
        subprocess.run(['pkill', '-f', TBEARS_BLOCK_MANAGER], stdout=subprocess.DEVNULL)

        # stop iconservice
        subprocess.run(['iconservice', 'stop', '-c', f'{TBEARS_CLI_ENV}'], stdout=subprocess.DEVNULL)

        time.sleep(2)

        print(f'Stopped tbears service successfully')

    def check_command(self, command):
        return hasattr(self, command)

    @staticmethod
    def _start_iconservice(conf: dict, config_path: str):
        subprocess.run(['iconservice', 'start', '-c', f'{config_path}'], stdout=subprocess.DEVNULL)

    @staticmethod
    def _start_iconrpcserver(conf: dict, config_path: str):
        subprocess.run(['iconrpcserver', 'start', '-tbears', '-c', f'{config_path}'], stdout=subprocess.DEVNULL)

    @staticmethod
    def _start_blockmanager(conf: dict):
        # make params
        params = {'-ch': conf.get(ConfigKey.CHANNEL, None),
                  '-at': conf.get(ConfigKey.AMQP_TARGET, None),
                  '-ak': conf.get(ConfigKey.AMQP_KEY, None),
                  '-c': conf.get('config', None)}

        custom_argv = []
        for k, v in params.items():
            if v:
                custom_argv.append(k)
                custom_argv.append(v)

        # Run block_manager in background mode
        subprocess.Popen([sys.executable, '-m', BLOCKMANAGER_MODULE_NAME, *custom_argv], close_fds=True)

    @staticmethod
    def is_service_running(name: str = TBEARS_BLOCK_MANAGER) -> bool:
        """ Check if server is running.
        :return: True or False
        """
        cmdlines = CommandServer._get_process_command_list(name.encode('utf-8'))
        if cmdlines:
            return True
        return False

    @staticmethod
    def _get_process_command_list(prefix: bytes) -> list:
        if os.path.exists('/proc'):
            cmd_lines = []
            pids = [pid for pid in os.listdir('/proc') if pid.isdigit()]
            for pid in pids:
                try:
                    cmdpath = os.path.join('/proc', pid, 'cmdline')
                    with open(cmdpath, 'rb') as fd:
                        cmdline = fd.read().rstrip(b'\x00')
                        if cmdline.startswith(prefix):
                            cmd_lines.append(cmdline.decode())
                except IOError:
                    continue
        else:
            result = subprocess.run(['ps', '-eo', 'command'], stdout=subprocess.PIPE)
            cmd_lines = [cmdline.decode().rstrip()
                         for cmdline in result.stdout.split(b'\n')
                         if cmdline.startswith(prefix)]
        return cmd_lines

    @staticmethod
    def is_port_available(conf):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # if socket is connected, the result code is 0 (false).
        result = sock.connect_ex(('127.0.0.1', conf['port']))
        sock.close()
        return result != 0

    @staticmethod
    def write_server_conf(conf: dict):
        write_conf = {
            "hostAddress": conf['hostAddress'],
            "port": conf['port'],
            "scoreRootPath": conf['scoreRootPath'],
            "stateDbRootPath": conf['stateDbRootPath'],
            ConfigKey.CHANNEL: conf.get(ConfigKey.CHANNEL, None),           # to stop iconservice
            ConfigKey.AMQP_TARGET: conf.get(ConfigKey.AMQP_TARGET, None),   # to stop iconservice
            ConfigKey.AMQP_KEY: conf.get(ConfigKey.AMQP_KEY, None)          # to stop iconservice
        }
        Logger.debug(f"Write server Info.({conf}) to {TBEARS_CLI_ENV}", TBEARS_CLI_TAG)
        file_path = TBEARS_CLI_ENV
        file_name = file_path[file_path.rfind('/') + 1:]
        parent_directory = file_path[:file_path.rfind('/')]
        try:
            write_file(parent_directory=parent_directory, file_name=file_name, contents=json.dumps(write_conf),
                       overwrite=True)
        except Exception as e:
            print(f"Can't write conf to file. {e}")
        except TBearsWriteFileException as e:
            print(f"{e}")

    @staticmethod
    def get_server_conf(file_path: str = TBEARS_CLI_ENV) -> Optional[dict]:
        try:
            with open(f'{file_path}') as f:
                conf = json.load(f)
        except Exception as e:
            Logger.debug(f"Can't read server configuration({file_path}. {e}")
            return None

        Logger.debug(f"Get server info {conf}", TBEARS_CLI_TAG)
        return conf

    @staticmethod
    def _delete_server_conf() -> None:
        if os.path.exists(TBEARS_CLI_ENV):
            os.remove(TBEARS_CLI_ENV)



import re
from urllib import parse


def uri_parser(uri: str) -> (str, int):

    uri = parse.urlparse(uri)
    _uri = uri.scheme + '://' + uri.netloc
    _version = re.search(r'(?<=\bv)\d+', uri.path).group(0)

    return _uri, int(_version)



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys
import time
from asyncio import get_event_loop
from copy import deepcopy
from typing import Union

import setproctitle
from earlgrey import MessageQueueService
from iconcommons.icon_config import IconConfig
from iconcommons.logger import Logger
from iconservice.icon_constant import DATA_BYTE_ORDER, DEFAULT_BYTE_SIZE
from iconservice.utils.bloom import BloomFilter

from tbears.block_manager.block import Block
from tbears.block_manager.channel_service import ChannelService, ChannelTxCreatorService
from tbears.block_manager.hash_utils import generate_hash
from tbears.block_manager.icon_service import IconStub
from tbears.block_manager.task import Periodic, Immediate
from tbears.config.tbears_config import ConfigKey, tbears_server_config, keystore_test1
from tbears.util import create_hash, get_tbears_version

TBEARS_BLOCK_MANAGER = 'tbears_block_manager'

ICON_SCORE_QUEUE_NAME_FORMAT = "IconScore.{channel_name}.{amqp_key}"
CHANNEL_QUEUE_NAME_FORMAT = "Channel.{channel_name}.{amqp_key}"
CHANNEL_TX_CREATOR_QUEUE_NAME_FORMAT = "ChannelTxCreator.{channel_name}.{amqp_key}"


class PRepManager(object):
    def __init__(self, is_generator_rotation: bool, gen_count_per_leader: int, prep_list: list = None):
        self._prep_list: list = deepcopy(prep_list)
        self._is_generator_rotation: bool = is_generator_rotation
        self._gen_count_per_leader: int = gen_count_per_leader
        self._gen_count: int = 0
        self._generator: str = keystore_test1['address']
        self._prev_generator: str = keystore_test1['address']
        self._prev_votes: list = []
        self.prev_preps: list = deepcopy(prep_list)

    def register_preps(self, data: dict):
        if data is None:
            return

        self._prep_list = data.get('preps', None)
        self._gen_count = 0

    @staticmethod
    def _create_prev_block_contributors_format(generator: str, validators: list = []) -> dict:
        validator_id_list = []
        for v in validators:
            if isinstance(v, dict) and 'id' in v:
                validator_id_list.append(v['id'])

        prev_block_contributors_format = \
            {
                "prevBlockGenerator": generator,
                "prevBlockValidators": validator_id_list
            }

        Logger.debug(f'prev_block_contributors_format: {prev_block_contributors_format}')
        return prev_block_contributors_format

    def get_prev_block_contributors_info(self) -> dict:
        if self.prep_list is None or len(self.prep_list) == 0:
            # set generator with test1 and validators with empty list
            prev_block_contributors_format = self._create_prev_block_contributors_format(keystore_test1['address'])
            return prev_block_contributors_format

        prev_block_contributors_format = self._create_prev_block_contributors_format(self._prep_list[0].get('id'),
                                                                                     self._prep_list[1:])
        # rotate leader after generate block 10 times
        self._prev_generator = self._generator
        if self._is_generator_rotation:
            self._gen_count += 1
            if self._gen_count == self._gen_count_per_leader:
                self._gen_count = 0
                self._prep_list.append(self._prep_list.pop(0))
                self._generator = self._prep_list[0]['id']
                Logger.debug(f"generator rotated. generator: {self._generator}", TBEARS_BLOCK_MANAGER)

        return prev_block_contributors_format

    @property
    def prev_generator(self):
        return self._prev_generator

    @property
    def generator(self):
        return self._generator

    @property
    def prep_list(self):
        return self._prep_list

    def set_prev_votes(self, block_height: int, block_hash: str, timestamp: int):
        self._prev_votes.clear()
        if self.prep_list:
            for index, prep in enumerate(self.prep_list):
                if index == 0:
                    self._prev_votes.append(None)
                else:
                    self._prev_votes.append({
                        "rep": prep.get("id"),
                        "timestamp": timestamp,
                        "blockHeight": block_height,
                        "blockHash": block_hash,
                        "signature": "tbears_block_manager_does_not_support_prev_vote_signature"
                    })

    @property
    def prev_votes(self):
        return self._prev_votes


class BlockManager(object):
    def __init__(self, conf: 'IconConfig'):
        self._conf = conf
        self._channel_mq_name = None
        self._tx_creator_mq_name = None
        self._icon_mq_name = None
        self._amqp_target = None
        self._channel_service = None
        self._tx_creator_service = None
        self._icon_stub = None
        self._block: 'Block' = Block(f'{conf["stateDbRootPath"]}/tbears')
        self._tx_queue = []
        self._prep_manager = PRepManager(
            is_generator_rotation=self._conf[ConfigKey.BLOCK_GENERATOR_ROTATION],
            gen_count_per_leader=self._conf[ConfigKey.BLOCK_GENERATE_COUNT_PER_LEADER])
        self._genesis_addr: str = self._conf["genesis"]["accounts"][0]["address"]
        self._test1_addr: str = self._conf["genesis"]["accounts"][2]["address"]

    @property
    def block(self) -> 'Block':
        return self._block

    def serve(self):
        async def _serve():
            try:
                await self.init()
            except RuntimeError as e:
                msg = f'Failed to connect to MQ. Check rabbitMQ service. ({e})'
                Logger.error(msg, TBEARS_BLOCK_MANAGER)
                print(msg)
                self.close()
                # TODO how to notify process status to parent process or system
                return

            Logger.info(f'tbears block_manager service started!', TBEARS_BLOCK_MANAGER)

        channel = self._conf[ConfigKey.CHANNEL]
        amqp_key = self._conf[ConfigKey.AMQP_KEY]
        amqp_target = self._conf[ConfigKey.AMQP_TARGET]

        self._channel_mq_name = CHANNEL_QUEUE_NAME_FORMAT.format(channel_name=channel, amqp_key=amqp_key)
        self._tx_creator_mq_name = CHANNEL_TX_CREATOR_QUEUE_NAME_FORMAT.format(channel_name=channel, amqp_key=amqp_key)
        self._icon_mq_name = ICON_SCORE_QUEUE_NAME_FORMAT.format(channel_name=channel, amqp_key=amqp_key)
        self._amqp_target = amqp_target

        Logger.info(f'==========tbears block_manager params==========', TBEARS_BLOCK_MANAGER)
        Logger.info(f'amqp_target : {amqp_target}', TBEARS_BLOCK_MANAGER)
        Logger.info(f'amqp_key    : {amqp_key}', TBEARS_BLOCK_MANAGER)
        Logger.info(f'queue_name  : {self._channel_mq_name}', TBEARS_BLOCK_MANAGER)
        Logger.info(f'            : {self._tx_creator_mq_name}', TBEARS_BLOCK_MANAGER)
        Logger.info(f'            : {self._icon_mq_name}', TBEARS_BLOCK_MANAGER)
        Logger.info(f'==========tbears block_manager params==========', TBEARS_BLOCK_MANAGER)

        # start message queue service
        loop = MessageQueueService.loop
        loop.create_task(_serve())
        loop.run_forever()

    async def init(self):
        """
        Initialize tbears block_manager
        :return:
        """
        Logger.debug(f'Initialize started!!', TBEARS_BLOCK_MANAGER)

        await self._init_channel()
        await self._init_tx_creator()
        await self._init_icon()
        if self._conf[ConfigKey.BLOCK_MANUAL_CONFIRM]:
            await self._init_immediate()
        else:
            await self._init_periodic()

        Logger.debug(f'Initialize done!!', TBEARS_BLOCK_MANAGER)

    async def _init_channel(self):
        """
        Initialize 'Channel' message queue
        :return:
        """
        Logger.debug(f'Initialize channel started!!', TBEARS_BLOCK_MANAGER)

        self._channel_service = ChannelService(self._amqp_target, self._channel_mq_name,
                                               block_manager=self)

        await self._channel_service.connect(exclusive=True)

        Logger.debug(f'Initialize channel done!!', TBEARS_BLOCK_MANAGER)

    async def _init_tx_creator(self):
        """
        Initialize 'ChannelTxCreator' message queue
        :return:
        """
        Logger.debug(f'Initialize tx creator started!!', TBEARS_BLOCK_MANAGER)

        self._tx_creator_service = ChannelTxCreatorService(self._amqp_target, self._tx_creator_mq_name,
                                                           block_manager=self)

        await self._tx_creator_service.connect(exclusive=True)

        Logger.debug(f'Initialize tx creator done!!', TBEARS_BLOCK_MANAGER)

    async def _init_icon(self):
        """
        Initialize 'IconScore' message queue and load genesis block if needed
        :return:
        """
        Logger.debug(f'Initialize ICON started!!', TBEARS_BLOCK_MANAGER)

        # make MQ stub
        self._icon_stub = IconStub(amqp_target=self._amqp_target, route_key=self._icon_mq_name)

        await self._icon_stub.connect()
        await self._icon_stub.async_task().hello()

        # query and set P-Rep list
        response = await self._icon_stub.async_task().call({"method": "ise_getPRepList"})
        if "result" in response:
            self._prep_manager.register_preps(response["result"])

        # send genesis block
        if self.block.block_height != -1:
            Logger.debug(f'Initialize ICON done!! block_height: {self.block.block_height}', TBEARS_BLOCK_MANAGER)
            return None

        tx_hash = create_hash(b'genesis')
        tx_timestamp_us = int(time.time() * 10 ** 6)
        request_params = {'txHash': tx_hash, 'timestamp': hex(tx_timestamp_us)}

        tx = {
            'method': '',
            'params': request_params,
            'genesisData': self._conf['genesis']
        }

        request = {'transactions': [tx]}
        block_height: int = self.block.block_height + 1
        block_timestamp_us = tx_timestamp_us
        block_hash = create_hash(block_timestamp_us.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER))

        request['block'] = {
            'blockHeight': hex(block_height),
            'blockHash': block_hash,
            'timestamp': hex(block_timestamp_us)
        }

        response = await self._icon_stub.async_task().invoke(request)

        tx_results = response['txResults']

        precommit_request = {'blockHeight': hex(block_height),
                             'oldBlockHash': block_hash,
                             'newBlockHash': block_hash}
        await self._icon_stub.async_task().write_precommit_state(precommit_request)

        tx_result = tx_results[tx_hash]
        tx_result['from'] = request_params.get('from', '')
        # tx_result['txHash'] must start with '0x'
        # tx_hash must not start with '0x'
        if tx_hash[:2] != '0x':
            tx_result['txHash'] = f'0x{tx_hash}'
        else:
            tx_result['txHash'] = tx_hash
            tx_hash = tx_hash[2:]

        # save transaction result
        self.block.save_txresult(tx_hash, tx_result)

        # save block
        block = self._make_block_data(block_hash, self._conf['genesis'], block_timestamp_us, response)
        self.block.save_block(block)

        # update block information
        self.block.commit_block(prev_block_hash=block_hash)

        Logger.debug(f'Initialize ICON done!! Load genesis block. block_height: {self.block.block_height}',
                     TBEARS_BLOCK_MANAGER)

    async def _init_periodic(self):
        """
        Initialize periodic task.
         - block confirmation
        :return:
        """
        Logger.debug(f'Initialize periodic task started!!', TBEARS_BLOCK_MANAGER)

        self.periodic = Periodic(func=self.process_block_data, interval=self._conf[ConfigKey.BLOCK_CONFIRM_INTERVAL])
        await self.periodic.start()

        Logger.debug(f'Initialize periodic task done!!', TBEARS_BLOCK_MANAGER)

    async def _init_immediate(self):
        """
        Initialize immediate task.
        :return:
        """
        Logger.debug(f'Initialize immediate task started!!', TBEARS_BLOCK_MANAGER)

        self.immediate = Immediate()
        await self.immediate.start()

        Logger.debug(f'Initialize periodic task done!!', TBEARS_BLOCK_MANAGER)

    def close(self):
        Logger.debug(f'close {TBEARS_BLOCK_MANAGER}', TBEARS_BLOCK_MANAGER)
        get_event_loop().stop()

    def add_tx(self, tx_hash: str, tx: dict):
        """
        Add transactions to queue for block confirmation
        :param tx_hash: transaction hash
        :param tx: transaction
        :return:
        """
        if self._conf[ConfigKey.BLOCK_MANUAL_CONFIRM] and self._check_debug_tx(tx):
            self.immediate.add_func(func=self.process_block_data)
        else:
            tx_copy = deepcopy(tx)

            # add txHash
            tx_copy['txHash'] = tx_hash

            self._tx_queue.append(tx_copy)
            Logger.debug(f'Append tx to tx_queue: {self._tx_queue}', TBEARS_BLOCK_MANAGER)

    def _check_debug_tx(self, tx: dict) -> bool:
        if tx['from'] != self._test1_addr:
            return False

        if tx['to'] != self._genesis_addr:
            return False

        return True

    @property
    def tx_queue(self) -> list:
        """
        Get transaction queue
        :return:
        """
        return self._tx_queue

    def clear_tx(self) -> list:
        """
        return transaction queue and clear
        :return: transaction queue
        """
        tx_queue: list = self._tx_queue
        self._tx_queue = []

        return tx_queue

    async def process_block_data(self):
        """
        Process block data. Invoke block and save transactions, transaction results and block. Update block height and previous block hash.
        :return:
        """
        Logger.debug(f'process_block_data started!!', TBEARS_BLOCK_MANAGER)

        # clear tx_queue
        tx_list = self.clear_tx()

        if len(tx_list) == 0:
            if self._conf[ConfigKey.BLOCK_CONFIRM_EMPTY]:
                Logger.debug(f'Confirm empty block', TBEARS_BLOCK_MANAGER)
            else:
                Logger.debug(f'There are no transactions for block confirm. Bye~', TBEARS_BLOCK_MANAGER)
                return

        # make block hash. tbears block_manager is dev util
        block_timestamp_us = int(time.time() * 10 ** 6)
        block_hash = create_hash(block_timestamp_us.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER))

        # send invoke message to ICON
        prev_block_timestamp = block_timestamp_us - self._conf.get(ConfigKey.BLOCK_CONFIRM_INTERVAL, 0)
        self._prep_manager.set_prev_votes(self.block.block_height, self.block.prev_block_hash,
                                          prev_block_timestamp)
        response = await self._invoke_block(tx_list=tx_list, block_hash=block_hash, block_timestamp=block_timestamp_us)
        if response is None:
            Logger.debug(f'iconservice response None for invoke request.', TBEARS_BLOCK_MANAGER)
            return

        # send write precommit message and confirm block
        await self._confirm_block(tx_list=tx_list, invoke_response=response, block_hash=block_hash,
                                  timestamp=block_timestamp_us)
        Logger.debug(f'process_block_data done!!', TBEARS_BLOCK_MANAGER)

    async def _invoke_block(self, tx_list: list, block_hash: str, block_timestamp) -> dict:
        """
        Invoke block. Send 'invoke' message to iconservice and get response
        :param tx_list: transaction list
        :param block_hash: block hash
        :param block_timestamp: block confirm timestamp
        :return:
        """
        Logger.debug(f'invoke block start', TBEARS_BLOCK_MANAGER)
        block_height = self.block.block_height + 1
        prev_block_hash = self.block.prev_block_hash

        transactions = []
        for tx in tx_list:
            transaction = {
                "method": 'icx_sendTransaction',
                "params": tx
            }
            transactions.append(transaction)

        request = {
            'block': {
                'blockHeight': hex(block_height),
                'blockHash': block_hash,
                'prevBlockHash': prev_block_hash,
                'timestamp': hex(block_timestamp)
            },
            'transactions': transactions,
            'isBlockEditable': "0x1"
        }

        # add prev block contributor Info.
        prev_block_contributors = self._prep_manager.get_prev_block_contributors_info()
        request.update(prev_block_contributors)

        # send invoke message to iconservice
        response = await self._icon_stub.async_task().invoke(request)

        if 'error' in response:
            Logger.debug(f'Get error response from iconservice: {response}!!', TBEARS_BLOCK_MANAGER)
            return None

        Logger.debug(f'invoke block done. response: {response}!!', TBEARS_BLOCK_MANAGER)
        return response

    async def _confirm_block(self, tx_list: list, invoke_response: dict, block_hash: str, timestamp: int):
        """
        Confirm block. Send 'write_precommit_state' message and save transaction, transaction result and block data
        :param tx_list: transaction list
        :param invoke_response: invoke response
        :param block_hash: old block hash
        :param timestamp: block timestamp
        :return:
        """
        Logger.debug(f'confirm block start. tx_list:{tx_list}, invoke_response:{invoke_response}', TBEARS_BLOCK_MANAGER)

        # Set new P-Rep list
        self._prep_manager.register_preps(invoke_response.get('prep'))

        # remake tx_list with addedTransactions in invoke_response
        added_transactions: dict = invoke_response.get('addedTransactions')
        if added_transactions is not None:
            for tx_hash, tx in added_transactions.items():
                tx['txHash'] = tx_hash
                index: int = -1
                # find txResult and index
                for i, tx_result in enumerate(invoke_response.get('txResults')):
                    if tx_result.get('txHash') == tx_hash:
                        index = i
                        break

                # insert addedTransaction to tx_list
                if index != -1:
                    tx_list.insert(index, tx)
                else:
                    Logger.error(f"Can't find txResult of addedTransaction({tx_hash}: {tx})")

        # recalculate block_hash. tbears block_manager is dev util
        block_timestamp_us = int(time.time() * 10 ** 6)
        new_block_hash = create_hash(block_timestamp_us.to_bytes(DEFAULT_BYTE_SIZE, DATA_BYTE_ORDER))

        block_height = self.block.block_height + 1

        # save transaction result
        if block_height == 0:
            self.block.save_txresults_legacy(tx_list=tx_list, results=invoke_response)
        else:
            self.block.save_txresults(tx_results=invoke_response.get('txResults'), new_block_hash=new_block_hash)

        # save transactions
        self.block.save_transactions(tx_list=tx_list, block_hash=new_block_hash)

        # save block
        block_data = self._make_block_data(new_block_hash, tx_list, timestamp, invoke_response)
        self.block.save_block(block_data)
        self._prep_manager.prev_preps = self._prep_manager.prep_list

        # update block information
        self.block.commit_block(prev_block_hash=new_block_hash)

        precommit_request = {'blockHeight': hex(block_height),
                             'oldBlockHash': block_hash,
                             'newBlockHash': new_block_hash}
        # send write_precommit_state message to iconservice

        await self._icon_stub.async_task().write_precommit_state(precommit_request)

        Logger.debug(f'confirm block done.', TBEARS_BLOCK_MANAGER)

    def _make_block_data(self, block_hash: str, tx: Union[list, dict], timestamp: int, invoke_response: dict):
        is_genesis = isinstance(tx, dict)
        tx_list = []
        tx_results = invoke_response['txResults']
        logs_bloom = BloomFilter()
        if is_genesis:
            tx_list.append(tx)
            tx_results = list(tx_results.values())
        else:
            tx_list = tx
            for tx_result in tx_results:
                if "logsBloom" in tx_result.keys():
                    logs_bloom = logs_bloom + int(tx_result['logsBloom'], 16)

        block_height = self.block.block_height + 1

        preps = [] if self._prep_manager.prep_list is None else self._prep_manager.prep_list
        preps = [prep['id'] for prep in preps]
        preps = [f'00{prep[2:]}'.encode() for prep in preps]
        prev_preps = [] if self._prep_manager.prev_preps is None else self._prep_manager.prev_preps
        prev_preps = [prep['id'] for prep in prev_preps]
        prev_preps = [f'00{prep[2:]}'.encode() for prep in prev_preps]

        transactions_hash = generate_hash(tx_list)
        receipts_hash = generate_hash(tx_results)
        pre_votes_hash = generate_hash(self._prep_manager.prev_votes, "icx_vote")
        reps_hash = generate_hash(prev_preps)
        next_reps_hash = generate_hash(preps)
        block = {
            "version": "tbears",
            "prevHash": self.block.prev_block_hash if not is_genesis else "",
            "transactionsHash": transactions_hash,
            "stateHash": invoke_response['stateRootHash'],
            "receiptsHash": receipts_hash,
            "repsHash": reps_hash,
            "nextRepsHash": next_reps_hash,
            "leaderVotesHash": '0'*64,
            "prevVotesHash": pre_votes_hash,
            "logsBloom": hex(logs_bloom.value),
            "timestamp": timestamp,
            "transactions": tx_list,
            "leaderVotes": [],
            "prevVotes": self._prep_manager.prev_votes,
            "hash": block_hash,
            "height": block_height,
            "leader": self._prep_manager.prev_generator,
            "signature": "tbears_block_manager_does_not_support_block_signature" if not is_genesis else "",
            "nextLeader": self._prep_manager.generator
        }
        return block


def create_parser():
    """
    Create tbears_block_manager argument parser
    :return:
    """
    parser = argparse.ArgumentParser(prog=TBEARS_BLOCK_MANAGER,
                                     description=f'{TBEARS_BLOCK_MANAGER} v{get_tbears_version()} arguments')
    parser.add_argument('-ch', dest=ConfigKey.CHANNEL, help='Message Queue channel')
    parser.add_argument('-at', dest=ConfigKey.AMQP_TARGET, help='AMQP target info')
    parser.add_argument('-ak', dest=ConfigKey.AMQP_KEY,
                        help="Key sharing peer group using queue name. \
                        Use it if more than one peer connect to a single MQ")
    parser.add_argument('-bi', '--block-confirm-interval', dest=ConfigKey.BLOCK_CONFIRM_INTERVAL, type=int,
                        help='Block confirm interval in second')
    parser.add_argument('-be', '--block-confirm-empty', dest=ConfigKey.BLOCK_CONFIRM_EMPTY, type=bool,
                        help='Confirm empty block')
    parser.add_argument('-c', '--config', help='Configuration file path')

    return parser


def main():
    # Parse arguments.
    try:
        parser = create_parser()
    except Exception as e:
        exit(f"Argument parsing exception : {e}")

    args = parser.parse_args(sys.argv[1:])

    if args.config:
        conf_path = args.config
        if not IconConfig.valid_conf_path(conf_path):
            print(f'Invalid configuration file : {conf_path}')
            sys.exit(1)
    else:
        conf_path = str()

    # Load configuration
    conf = IconConfig(conf_path, tbears_server_config)
    conf.load()
    conf.update_conf(dict(vars(args)))
    Logger.load_config(conf)
    Logger.print_config(conf, TBEARS_BLOCK_MANAGER)

    setproctitle.setproctitle(f'{TBEARS_BLOCK_MANAGER}.{conf[ConfigKey.CHANNEL]}.{conf[ConfigKey.AMQP_KEY]}')

    # run block_manager service
    block_manager = BlockManager(conf=conf)
    block_manager.serve()

    Logger.info('===============tbears block_manager done================', TBEARS_BLOCK_MANAGER)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

with open('requirements.txt') as requirements:
    requires = list(requirements)

version = os.environ.get('VERSION')
if version is None:
    with open(os.path.join('.', 'VERSION')) as version_file:
        version = version_file.read().strip()

setup_options = {
    'name': 'tbears',
    'version': version,
    'description': 'Test suite for ICON SCORE development',
    'long_description': open('README.md').read(),
    'long_description_content_type': "text/markdown",
    'url': 'https://github.com/icon-project/t-bears',
    'author': 'ICON Foundation',
    'author_email': 'foo@icon.foundation',
    'packages': find_packages(exclude=['tests*', 'docs']),
    'include_package_data': True,
    'py_modules': ['tbears'],
    'license': "Apache License 2.0",
    'install_requires': requires,
    'test_suite': 'tests',
    'entry_points': {
        'console_scripts': [
            'tbears=tbears:main'
        ],
    },
    'classifiers': [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
}

setup(**setup_options)



# -*- coding: utf-8 -*-

# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING

from iconservice.base.exception import InvalidRequestException

from .db import PlyvelDB

if TYPE_CHECKING:
    from iconservice.base.address import Address
    from iconservice.iconscore.icon_score_context import IconScoreContext


class IcxEngine:
    db: PlyvelDB = None

    @classmethod
    def get_balance(cls, context: 'IconScoreContext', address: 'Address') -> int:
        balance = none_to_zero(cls.db.get(context, address.to_bytes()))
        return balance

    @classmethod
    def transfer(cls, context: 'IconScoreContext', _from: 'Address', _to: 'Address', amount: int):
        sender_address = _from.to_bytes()
        receiver_address = _to.to_bytes()

        sender_balance = none_to_zero(cls.db.get(context, sender_address))
        receiver_balance = none_to_zero(cls.db.get(context, receiver_address))

        if sender_balance < amount:
            raise InvalidRequestException('out of balance')

        sender_balance -= amount
        receiver_balance += amount

        cls.db._db[sender_address] = sender_balance
        cls.db._db[receiver_address] = receiver_balance


def none_to_zero(balance) -> int:
    if balance is None:
        return 0
    return balance



# Copyright 2019 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib

from iconcommons import Logger
from secp256k1 import PublicKey, ALL_FLAGS

_public_key = PublicKey(flags=ALL_FLAGS)


def address_from_pubkey(pubkey: bytes):
    hash_pub = hashlib.sha3_256(pubkey[1:]).hexdigest()
    return f"hx{hash_pub[-40:]}"


def verify_signature(msg_hash: bytes, signature: bytes, sender: str) -> bool:
    if isinstance(msg_hash, bytes) \
            and len(msg_hash) == 32 \
            and isinstance(signature, bytes) \
            and len(signature) == 65:

        origin_sig, rec_id = signature[:-1], signature[-1]
        recoverable_sig = _public_key.ecdsa_recoverable_deserialize(origin_sig, rec_id)
        internal_pubkey = _public_key.ecdsa_recover(msg_hash,
                                                    recoverable_sig,
                                                    raw=True, digest=None)
        public_key = PublicKey(internal_pubkey,
                               raw=False,
                               ctx=_public_key.ctx).serialize(compressed=False)

        address: str = address_from_pubkey(public_key)
        Logger.info(f'Expected address={sender}', "verify_signature")
        Logger.info(f'Signed address={address}', "verify_signature")
        if address == sender:
            return True

    return False



# Copyright 2017 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" A message class for the loopchain """

class Request:
    status = 1
    is_alive = 2
    stop = -9

    peer_peer_list = 600
    peer_get_leader = 601  # get leader peer object
    peer_complain_leader = 602  # complain leader peer is no response
    peer_reconnect_to_rs = 603  # reconnect to rs when rs restart detected.
    peer_restart_channel = 604

    rs_get_configuration = 800
    rs_set_configuration = 801
    rs_send_channel_manage_info_to_rs = 802
    rs_restart_channel = 803
    rs_delete_peer = 804

    tx_connect_to_leader = 901  # connect to leader
    tx_connect_to_inner_peer = 902  # connect to mother peer service in same inner gRPC micro service network
    get_tx_result = 903  # json-rpc:icx_getTransactionResult
    get_balance = 905  # josn-rpc:icx_getBalance
    get_tx_by_address = 906  # json-rpc:icx_getTransactionByAddress
    get_total_supply = 907  # json-rpc:icx_getTotalSupply


class MetaParams:
    class ScoreLoad:
        repository_path = "repository_path"
        score_package = "score_package"
        base = "base"
        peer_id = "peer_id"

    class ScoreInfo:
        score_id = "score_id"
        score_version = "score_version"


class Response:
    success = 0
    success_validate_block = 1
    success_announce_block = 2
    fail = -1
    fail_validate_block = -2
    fail_announce_block = -3
    fail_wrong_block_hash = -4
    fail_no_leader_peer = -5
    fail_validate_params = -6
    fail_made_block_count_limited = -7
    fail_wrong_subscribe_info = -8
    fail_connect_to_leader = -9
    fail_add_tx_to_leader = -10
    fail_create_tx = -11
    fail_invalid_peer_target = -12
    fail_not_enough_data = -13
    fail_tx_pre_validate = -14
    fail_subscribe_limit = -15
    fail_invalid_key_error = -16
    fail_wrong_block_height = -17
    fail_tx_invalid_unknown = -100
    fail_tx_invalid_hash_format = -101
    fail_tx_invalid_hash_generation = -102
    fail_tx_invalid_hash_not_match = -103
    fail_tx_invalid_address_not_match = -104
    fail_tx_invalid_address_format = -105
    fail_tx_invalid_signature = -106
    fail_tx_invalid_params = -107
    fail_tx_invalid_duplicated_hash = -108
    fail_tx_invalid_out_of_time_bound = -109
    fail_tx_invalid_wrong_nid = -110
    fail_tx_not_invoked = -111

    fail_no_peer_info_in_rs = -800
    timeout_exceed = -900
    not_treat_message_code = -999
    fail_illegal_params = -1000


responseCodeMap = {
    Response.success:
        (Response.success,"success"),

    Response.success_validate_block:
        (Response.success_validate_block, "success validate block"),

    Response.success_announce_block:
        (Response.success_announce_block, "success announce block"),

    Response.fail:
        (Response.fail, "fail"),

    Response.fail_validate_block:
        (Response.fail_validate_block, "fail validate block"),

    Response.fail_announce_block:
        (Response.fail_announce_block, "fail announce block"),

    Response.fail_wrong_block_hash:
        (Response.fail_wrong_block_hash, "fail wrong block hash"),

    Response.fail_no_leader_peer:
        (Response.fail_no_leader_peer, "fail no leader peer"),

    Response.fail_validate_params:
        (Response.fail_validate_params, "fail validate params"),

    Response.fail_wrong_subscribe_info:
        (Response.fail_wrong_subscribe_info, "fail wrong subscribe info"),

    Response.fail_connect_to_leader:
        (Response.fail_connect_to_leader, "fail connect to leader"),

    Response.fail_add_tx_to_leader:
        (Response.fail_add_tx_to_leader, "fail add tx to leader"),

    Response.fail_invalid_peer_target:
        (Response.fail_invalid_peer_target, "fail invalid peer target for channel"),

    Response.fail_not_enough_data:
        (Response.fail_not_enough_data, "fail not enough data"),

    Response.fail_tx_pre_validate:
        (Response.fail_tx_pre_validate, "fail tx pre-validate"),

    Response.fail_subscribe_limit:
        (Response.fail_subscribe_limit, "fail subscribe limit"),

    Response.fail_no_peer_info_in_rs:
        (Response.fail_no_peer_info_in_rs, "fail no peer info in radio station"),

    Response.fail_create_tx:
        (Response.fail_create_tx, "fail create tx to peer"),

    Response.fail_wrong_block_height:
        (Response.fail_wrong_block_height, "fail wrong block height"),

    Response.fail_tx_invalid_unknown:
        (Response.fail_tx_invalid_unknown, "fail tx invalid unknown"),

    Response.fail_tx_invalid_hash_format:
        (Response.fail_tx_invalid_hash_format, "fail tx invalid hash format"),

    Response.fail_tx_invalid_hash_generation:
        (Response.fail_tx_invalid_hash_generation, "fail tx invalid hash generation"),

    Response.fail_tx_invalid_address_not_match:
        (Response.fail_tx_invalid_address_not_match, "fail tx invalid address not match"),

    Response.fail_tx_invalid_address_format:
        (Response.fail_tx_invalid_address_format, "fail tx invalid address"),

    Response.fail_tx_invalid_hash_not_match:
        (Response.fail_tx_invalid_hash_not_match, "fail tx invalid hash not match"),

    Response.fail_tx_invalid_signature:
        (Response.fail_tx_invalid_signature, "fail tx invalid signature"),

    Response.fail_tx_invalid_params:
        (Response.fail_tx_invalid_params, "fail tx invalid params"),

    Response.fail_tx_invalid_duplicated_hash:
        (Response.fail_tx_invalid_duplicated_hash, "fail tx invalid duplicated hash"),

    Response.fail_tx_invalid_out_of_time_bound:
        (Response.fail_tx_invalid_out_of_time_bound, "fail tx invalid out of time bound"),

    Response.fail_tx_invalid_wrong_nid:
        (Response.fail_tx_invalid_wrong_nid, "fail tx invalid no nid"),

    Response.fail_tx_not_invoked:
        (Response.fail_tx_not_invoked, "Pending transaction"),

    Response.timeout_exceed:
        (Response.timeout_exceed, "timeout exceed"),

    Response.fail_illegal_params:
        (Response.fail_illegal_params, "fail_illegal_params")
}


def get_response_code(code):
    return responseCodeMap[code][0]


def get_response_msg(code):
    return responseCodeMap[code][1]


def get_response(code):
    return responseCodeMap[code][0], responseCodeMap[code][1]



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from argparse import ArgumentTypeError

from iconservice.base.address import is_icon_address_valid

from tbears.util import is_valid_hash


class IconPath(str):
    def __init__(self, mode: str = 'r'):
        self._mode = mode

    def __call__(self, string: str) -> str:
        if self._mode == 'r' and not os.path.exists(string):
            raise ArgumentTypeError(f"There is no '{string}'")
        elif self._mode == 'w' and os.path.exists(string):
            raise ArgumentTypeError(f"'{string}' must be empty")
        elif self._mode == 'd' and not os.path.isdir(string):
            raise ArgumentTypeError(f"There is no directory '{string}'")

        if os.path.isdir(string) and string.endswith('/'):
            return string[:-1]

        return string


class IconAddress(str):
    def __init__(self, prefix: str = 'all'):
        self._prefix = prefix

    def __call__(self, string: str) -> str:
        # check prefix of given address (string). if not 'cx' or 'hx', raise error
        if not is_icon_address_valid(string):
            raise ArgumentTypeError(f"Invalid address '{string}'")

        if self._prefix != 'all':
            if self._prefix != string[:2]:
                raise ArgumentTypeError(f"Invalid address '{string}'. Address must start with '{self._prefix}'")

        return string


def hash_type(string: str) -> str:
    # check hash's length, prefix, lowercase.
    if not is_valid_hash(string):
        raise ArgumentTypeError(f"Invalid hash '{string}'")

    return string


def port_type(string: str) -> int:
    try:
        port = int(string, 10)
    except ValueError:
        raise ArgumentTypeError(f"Invalid integer value '{string}'")
    except TypeError as e:
        raise ArgumentTypeError(f'Invalid type. {e}')

    if port < 0 or port > 65535:
        raise ArgumentTypeError(f"Invalid port '{string}'. Port must be 0 < port < 65536")

    return port


def non_negative_num_type(string: str) -> str:
    try:
        value = int(string, 10)
    except ValueError:
        try:
            value = int(string, 16)
        except ValueError:
            raise ArgumentTypeError(f"Invalid integer value '{string}'. Hexadecimal and decimal values are allowed")
    except TypeError as e:
        raise ArgumentTypeError(f'Invalid type. {e}')

    if value < 0:
        raise ArgumentTypeError(f"Invalid non-negative number '{value}'")

    return hex(value)


def loop(input_value: str) -> int:
    try:
        value_in_float = float(input_value)
    except ValueError:
        raise ArgumentTypeError(f'You entered invalid value {input_value}')
    if value_in_float != float(int(value_in_float)):
        raise ArgumentTypeError(f'You entered invalid value {input_value}')

    value_in_str = input_value.lower()
    if value_in_str.find("e") == -1:
        return int(value_in_str, 10)
    else:
        significand, exponent = value_in_str.split("e")
        return int(significand) * 10 ** int(exponent)



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from earlgrey import MessageQueueStub, message_queue_task
from iconcommons.logger import Logger


if TYPE_CHECKING:
    from earlgrey import RobustConnection


class IconScoreInnerTask(object):
    """
    Send request to 'iconscore' message queue
    """
    @message_queue_task
    async def hello(self):
        pass

    @message_queue_task
    async def invoke(self, request: dict) -> dict:
        pass

    @message_queue_task
    async def query(self, request: dict) -> dict:
        pass

    @message_queue_task
    async def write_precommit_state(self, request: dict) -> dict:
        pass

    @message_queue_task
    async def remove_precommit_state(self, request: dict) -> dict:
        pass

    @message_queue_task
    async def call(self, request: dict) -> dict:
        pass


class IconStub(MessageQueueStub[IconScoreInnerTask]):
    TaskType = IconScoreInnerTask

    def _callback_connection_lost_callback(self, connection: 'RobustConnection'):
        Logger.error(f'IconStub lost message queue connection', 'tbears_block_manager')
        self._task._block_manager.close()
